import uuid
from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.ponos.models import Farm
from arkindex.process.models import Process, ProcessDatasetSet, ProcessMode
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Dataset, DatasetSet
from arkindex.users.models import Role, User

# Using the fake DB fixtures creation date when needed
FAKE_CREATED = "2020-02-02T01:23:45.678000Z"


class TestProcessDatasetSets(FixtureAPITestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name="Private corpus")
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = FAKE_CREATED
            cls.private_dataset = cls.private_corpus.datasets.create(
                name="Dead sea scrolls",
                description="Human instrumentality manual",
                creator=cls.user
            )
        DatasetSet.objects.bulk_create([
            DatasetSet(dataset_id=cls.private_dataset.id, name=set_name)
            for set_name in ["dev", "train", "test"]
        ])
        cls.test_user = User.objects.create(email="katsuragi@nerv.co.jp", verified_email=True)
        cls.private_corpus.memberships.create(user=cls.test_user, level=Role.Admin.value)

        # Datasets from another corpus
        cls.dataset1, cls.dataset2 = Dataset.objects.filter(corpus=cls.corpus).order_by("name")

        cls.dataset_process = Process.objects.create(
            creator_id=cls.user.id,
            mode=ProcessMode.Dataset,
            corpus_id=cls.private_corpus.id,
            farm=Farm.objects.get(name="Wheat farm")
        )
        ProcessDatasetSet.objects.bulk_create(
            ProcessDatasetSet(process=cls.dataset_process, set=dataset_set)
            for dataset_set in cls.dataset1.sets.all()
        )
        ProcessDatasetSet.objects.bulk_create(
            ProcessDatasetSet(process=cls.dataset_process, set=dataset_set)
            for dataset_set in cls.private_dataset.sets.all()
        )

        # Control process to check that its datasets are not retrieved
        cls.dataset_process_2 = Process.objects.create(
            creator_id=cls.user.id,
            mode=ProcessMode.Dataset,
            corpus_id=cls.corpus.id
        )
        ProcessDatasetSet.objects.bulk_create(
            ProcessDatasetSet(process=cls.dataset_process_2, set=dataset_set)
            for dataset_set in cls.dataset2.sets.all()
        )

    # List process dataset sets

    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:process-sets", kwargs={"pk": self.dataset_process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_process_does_not_exist(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:process-sets", kwargs={"pk": str(uuid.uuid4())}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch("arkindex.project.mixins.get_max_level", return_value=None)
    def test_list_process_access_level(self, get_max_level_mock):
        self.private_corpus.memberships.filter(user=self.test_user).delete()
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:process-sets", kwargs={"pk": self.dataset_process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have guest access to this process."})

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.test_user, self.private_corpus))

    def test_list(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:process-sets", kwargs={"pk": self.dataset_process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(response.json()["results"], [
            {
                "id": str(ProcessDatasetSet.objects.get(process=self.dataset_process, set=dataset_set).id),
                "dataset": {
                    "id": str(self.private_dataset.id),
                    "name": "Dead sea scrolls",
                    "description": "Human instrumentality manual",
                    "creator": "Test user",
                    "sets": [
                        {
                            "id": str(ds.id),
                            "name": ds.name
                        }
                        for ds in self.private_dataset.sets.order_by("name")
                    ],
                    "set_elements": None,
                    "corpus_id": str(self.private_corpus.id),
                    "state": "open",
                    "task_id": None,
                    "unique_elements": True,
                    "created": FAKE_CREATED,
                    "updated": FAKE_CREATED
                },
                "set_name": dataset_set.name
            }
            for dataset_set in self.private_dataset.sets.order_by("dataset__name", "name")
        ] + [
            {
                "id": str(ProcessDatasetSet.objects.get(process=self.dataset_process, set=dataset_set).id),
                "dataset": {
                    "id": str(self.dataset1.id),
                    "name": "First Dataset",
                    "description": "dataset number one",
                    "creator": "Test user",
                    "sets": [
                        {
                            "id": str(ds.id),
                            "name": ds.name
                        }
                        for ds in self.dataset1.sets.order_by("name")
                    ],
                    "set_elements": None,
                    "corpus_id": str(self.corpus.id),
                    "state": "open",
                    "task_id": None,
                    "unique_elements": True,
                    "created": FAKE_CREATED,
                    "updated": FAKE_CREATED
                },
                "set_name": dataset_set.name
            }
            for dataset_set in self.dataset1.sets.order_by("dataset__name", "name")
        ])

    # Create process dataset set

    def test_create_requires_login(self):
        test_set = self.dataset2.sets.get(name="test")
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": test_set.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_requires_verified(self):
        unverified_user = User.objects.create(email="email@mail.com")
        test_set = self.dataset2.sets.get(name="test")
        self.client.force_login(unverified_user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": test_set.id})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.get_max_level")
    def test_create_access_level(self, get_max_level_mock):
        cases = [None, Role.Guest.value, Role.Contributor.value]
        test_set = self.dataset2.sets.get(name="test")
        for level in cases:
            with self.subTest(level=level):
                get_max_level_mock.reset_mock()
                get_max_level_mock.return_value = level
                self.client.force_login(self.test_user)

                with self.assertNumQueries(3):
                    response = self.client.post(
                        reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": test_set.id})
                    )
                    self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

                self.assertEqual(response.json(), {"detail": "You do not have admin access to this process."})

                self.assertEqual(get_max_level_mock.call_count, 1)
                self.assertEqual(get_max_level_mock.call_args, call(self.test_user, self.private_corpus))

    def test_create_process_mode(self):
        cases = set(ProcessMode) - {ProcessMode.Dataset, ProcessMode.Local}
        test_set = self.dataset2.sets.get(name="test")
        for mode in cases:
            with self.subTest(mode=mode):
                self.dataset_process.mode = mode
                self.dataset_process.save()
                self.client.force_login(self.test_user)

                with self.assertNumQueries(6):
                    response = self.client.post(
                        reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": test_set.id})
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {"process": ['Dataset sets can only be added to or removed from processes of mode "dataset".']})

    def test_create_process_mode_local(self):
        self.client.force_login(self.user)
        test_set = self.dataset2.sets.get(name="test")
        local_process = Process.objects.get(creator=self.user, mode=ProcessMode.Local)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:process-set", kwargs={"process": local_process.id, "set": test_set.id})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "You do not have admin access to this process."})

    def test_create_wrong_process_uuid(self):
        self.client.force_login(self.test_user)
        test_set = self.dataset2.sets.get(name="test")
        wrong_id = uuid.uuid4()
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:process-set", kwargs={"process": wrong_id, "set": test_set.id})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"process": [f'Invalid pk "{str(wrong_id)}" - object does not exist.']})

    def test_create_wrong_set_uuid(self):
        self.client.force_login(self.test_user)
        wrong_id = uuid.uuid4()
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": wrong_id})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"set": [f'Invalid pk "{str(wrong_id)}" - object does not exist.']})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_create_dataset_access(self, filter_rights_mock):
        new_corpus = Corpus.objects.create(name="NERV")
        new_dataset = new_corpus.datasets.create(name="Eva series", description="We created the Evas from Adam", creator=self.user)
        test_set = new_dataset.sets.create(name="test")
        self.client.force_login(self.test_user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": test_set.id})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"set": [f'Invalid pk "{str(test_set.id)}" - object does not exist.']})

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.test_user, Corpus, Role.Guest.value))

    def test_create_unique(self):
        self.client.force_login(self.test_user)
        test_set = self.dataset1.sets.get(name="test")
        self.assertTrue(self.dataset_process.sets.filter(id=test_set.id).exists())

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": test_set.id})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"set": ["This dataset set is already selected in this process."]})

    def test_create_started(self):
        self.client.force_login(self.test_user)
        self.dataset_process.tasks.create(run=0, depth=0, slug="makrout")
        test_set = self.dataset2.sets.get(name="test")

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": test_set.id})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"process": ["Dataset sets cannot be updated on processes that have already started."]})

    def test_create(self):
        self.client.force_login(self.test_user)
        test_set = self.dataset2.sets.get(name="test")
        self.assertEqual(ProcessDatasetSet.objects.count(), 9)
        self.assertFalse(ProcessDatasetSet.objects.filter(process=self.dataset_process.id, set=test_set.id).exists())
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": test_set.id})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(ProcessDatasetSet.objects.count(), 10)
        self.assertTrue(ProcessDatasetSet.objects.filter(process=self.dataset_process.id, set=test_set.id).exists())
        self.assertQuerysetEqual(self.dataset_process.sets.order_by("dataset__name", "name"), [
            *self.private_dataset.sets.order_by("name"),
            *self.dataset1.sets.order_by("name"),
            test_set,
        ])
        created = ProcessDatasetSet.objects.get(process=self.dataset_process.id, set=test_set.id)
        self.assertDictEqual(response.json(), {
            "id": str(created.id),
            "dataset": {
                "id": str(self.dataset2.id),
                "name": "Second Dataset",
                "description": "dataset number two",
                "creator": "Test user",
                "sets": [
                    {
                        "id": str(ds.id),
                        "name": ds.name
                    }
                    for ds in self.dataset2.sets.all()
                ],
                "set_elements": None,
                "corpus_id": str(self.corpus.id),
                "state": "open",
                "task_id": None,
                "unique_elements": True,
                "created": FAKE_CREATED,
                "updated": FAKE_CREATED
            },
            "set_name": "test"
        })

    # Destroy process dataset set

    def test_destroy_requires_login(self):
        train_set = self.dataset1.sets.get(name="train")
        with self.assertNumQueries(0):
            response = self.client.delete(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": train_set.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_destroy_process_does_not_exist(self):
        train_set = self.dataset1.sets.get(name="train")
        self.client.force_login(self.test_user)
        wrong_id = uuid.uuid4()
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:process-set", kwargs={"process": wrong_id, "set": train_set.id})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"process": [f'Invalid pk "{str(wrong_id)}" - object does not exist.']})

    def test_destroy_set_does_not_exist(self):
        self.client.force_login(self.test_user)
        wrong_id = uuid.uuid4()
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": wrong_id})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"set": [f'Invalid pk "{str(wrong_id)}" - object does not exist.']})

    def test_destroy_not_found(self):
        train_set = self.dataset2.sets.get(name="train")
        self.assertFalse(ProcessDatasetSet.objects.filter(process=self.dataset_process, set=train_set).exists())
        self.client.force_login(self.test_user)
        with self.assertNumQueries(5):
            response = self.client.delete(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": train_set.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch("arkindex.project.mixins.get_max_level", return_value=None)
    def test_destroy_process_access_level(self, get_max_level_mock):
        train_set = self.dataset1.sets.get(name="train")
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": train_set.id})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have admin access to this process."})

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.test_user, self.private_corpus))

    def test_destroy_no_dataset_access_requirement(self):
        new_corpus = Corpus.objects.create(name="NERV")
        new_dataset = new_corpus.datasets.create(name="Eva series", description="We created the Evas from Adam", creator=self.user)
        test_set = new_dataset.sets.create(name="test")
        ProcessDatasetSet.objects.create(process=self.dataset_process, set=test_set)
        self.assertTrue(ProcessDatasetSet.objects.filter(process=self.dataset_process, set=test_set).exists())
        self.client.force_login(self.test_user)
        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": test_set.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(ProcessDatasetSet.objects.filter(process=self.dataset_process, set=test_set).exists())

    def test_destroy_process_mode(self):
        train_set = self.dataset1.sets.get(name="train")
        cases = set(ProcessMode) - {ProcessMode.Dataset, ProcessMode.Local}
        for mode in cases:
            with self.subTest(mode=mode):
                self.dataset_process.mode = mode
                self.dataset_process.save()
                self.client.force_login(self.test_user)

                with self.assertNumQueries(4):
                    response = self.client.delete(
                        reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": train_set.id}),
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {"process": ['Dataset sets can only be added to or removed from processes of mode "dataset".']})

    def test_destroy_process_mode_local(self):
        train_set = self.dataset1.sets.get(name="train")
        self.client.force_login(self.user)
        local_process = Process.objects.get(creator=self.user, mode=ProcessMode.Local)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:process-set", kwargs={"process": local_process.id, "set": train_set.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "You do not have admin access to this process."})

    def test_destroy_started(self):
        train_set = self.dataset1.sets.get(name="train")
        self.client.force_login(self.test_user)
        self.dataset_process.tasks.create(run=0, depth=0, slug="makrout")

        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": train_set.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"process": ["Dataset sets cannot be updated on processes that have already started."]})

    def test_destroy(self):
        train_set = self.dataset1.sets.get(name="train")
        self.assertTrue(ProcessDatasetSet.objects.filter(process=self.dataset_process, set=train_set).exists())
        self.client.force_login(self.test_user)
        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:process-set", kwargs={"process": self.dataset_process.id, "set": train_set.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(ProcessDatasetSet.objects.filter(process=self.dataset_process, set=train_set).exists())
