import uuid
from datetime import datetime, timezone
from unittest.mock import call, patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status
from rest_framework.exceptions import ValidationError

from arkindex.ponos.models import Agent, Farm, State
from arkindex.process.models import (
    FeatureUsage,
    GitRef,
    GitRefType,
    ProcessMode,
    Revision,
    Worker,
    WorkerRun,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersion, ModelVersionState
from arkindex.users.models import Role

ENV = {
    "ARKINDEX_PROCESS_ID": "12345"
}


class TestWorkerRuns(FixtureAPITestCase):
    """
    Test worker runs endpoints and methods
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.local_process = cls.user.processes.get(mode=ProcessMode.Local)
        cls.farm = Farm.objects.first()
        cls.process_1 = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
            farm=cls.farm,
        )
        cls.version_1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_1 = cls.version_1.worker
        cls.worker_custom = Worker.objects.get(slug="custom")
        cls.version_2 = WorkerVersion.objects.get(worker__slug="dla")
        cls.version_custom = cls.worker_custom.versions.get()
        cls.repo = cls.worker_1.repository
        cls.run_1 = cls.process_1.worker_runs.create(version=cls.version_1, parents=[])
        cls.run_custom = cls.local_process.worker_runs.get(version=cls.version_custom)
        cls.configuration_1 = cls.worker_1.configurations.create(name="My config", configuration={"key": "value"})
        worker_version = WorkerVersion.objects.exclude(worker=cls.version_1.worker).first()
        cls.configuration_2 = worker_version.worker.configurations.create(name="Config")
        cls.process_2 = cls.corpus.processes.create(creator=cls.user, mode=ProcessMode.Workers)
        # Add an execution access right on the worker
        cls.worker_1.memberships.create(user=cls.user, level=Role.Contributor.value)

        # Model and Model version setup
        cls.model_1 = Model.objects.create(name="My model")
        cls.model_1.memberships.create(user=cls.user, level=Role.Contributor.value)
        cls.model_version_1 = ModelVersion.objects.create(
            model=cls.model_1,
            state=ModelVersionState.Available,
            size=8,
            hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            archive_hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
        )

        cls.model_2 = Model.objects.create(name="Their model")
        cls.model_2.memberships.create(user=cls.user, level=Role.Guest.value)
        cls.model_version_2 = cls.model_2.versions.create(
            state=ModelVersionState.Available,
            tag="blah",
            size=8,
            hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            archive_hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
        )

        cls.model_3 = Model.objects.create(name="Our model", public=True)
        cls.model_version_3 = cls.model_3.versions.create(
            state=ModelVersionState.Available,
            tag="blah",
            size=8,
            hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            archive_hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
        )

        cls.agent = Agent.objects.create(
            farm=cls.farm,
            hostname="claude",
            cpu_cores=42,
            cpu_frequency=1e15,
            ram_total=99e9,
            last_ping=datetime.now(timezone.utc),
        )
        # Add custom attributes to make the agent usable as an authenticated user
        cls.agent.is_agent = True
        cls.agent.is_anonymous = False

    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:worker-run-list", kwargs={"pk": str(self.process_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_no_execution_right(self):
        """
        Worker runs attached to a process can be listed even if the user has no execution rights to workers
        This is due to the fact a user can see a process running on a corpus they have access
        """
        self.worker_1.memberships.all().delete()
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.get(reverse("api:worker-run-list", kwargs={"pk": str(self.process_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json()["count"], 1)

    def test_list(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.get(reverse("api:worker-run-list", kwargs={"pk": str(self.process_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["results"], [{
            "id": str(self.run_1.id),
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "parents": [],
            "model_version": None,
            "configuration": None,
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]}",
        }])

    def test_list_filter_process(self):
        run_2 = self.process_2.worker_runs.create(
            version=self.version_1,
            parents=[],
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.get(reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 1)
        self.assertEqual(data["results"][0]["id"], str(run_2.id))

    def test_create_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(self.version_1.id), "parents": []},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_create_no_execution_right(self, has_access_mock):
        """
        An execution access on the target worker version is required to create a worker run
        """
        self.worker_1.memberships.update(level=Role.Guest.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(self.version_1.id), "parents": []},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"worker_version_id": ["You do not have an execution access to this worker."]})

        self.assertListEqual(has_access_mock.call_args_list, [
            call(self.user, self.worker_1, Role.Contributor.value, skip_public=False),
            call(self.user, self.repo, Role.Contributor.value, skip_public=False),
        ])

    def test_create_invalid_version(self):
        self.client.force_login(self.user)
        version_id = uuid.uuid4()

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": version_id},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "worker_version_id": [f'Invalid pk "{version_id}" - object does not exist.'],
        })

    def test_create_unique(self):
        self.client.force_login(self.user)
        self.version_1.model_usage = FeatureUsage.Required
        self.version_1.save()
        cases = [
            (None, None),
            (None, self.configuration_1),
            (self.model_version_1, None),
            (self.model_version_1, self.configuration_1),
        ]
        for model_version, configuration in cases:
            with self.subTest(model_version=model_version, configuration=configuration):
                self.run_1.model_version = model_version
                self.run_1.configuration = configuration
                self.run_1.save()

                # Having a model version or a configuration adds one query for each
                query_count = 6 + bool(model_version) + bool(configuration)

                with self.assertNumQueries(query_count):
                    response = self.client.post(
                        reverse("api:worker-run-list", kwargs={"pk": str(self.process_1.id)}),
                        data={
                            "worker_version_id": str(self.version_1.id),
                            "model_version_id": str(model_version.id) if model_version else None,
                            "configuration_id": str(configuration.id) if configuration else None,
                        },
                        format="json",
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "__all__": ["A WorkerRun already exists on this process with the selected worker version, model version and configuration."],
                })

    def test_create_unavailable_version(self):
        self.version_1.state = WorkerVersionState.Error
        self.version_1.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(self.version_1.id), "parents": []},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"worker_version_id": ["This WorkerVersion is not in an Available state."]})

    def test_create_archived_worker(self):
        self.worker_1.archived = datetime.now(timezone.utc)
        self.worker_1.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(self.version_1.id), "parents": []},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"worker_version_id": ["This WorkerVersion is part of an archived worker."]})

    def test_create_archived_model(self):
        self.model_1.archived = datetime.now(timezone.utc)
        self.model_1.save()
        self.version_1.model_usage = FeatureUsage.Supported
        self.version_1.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(self.version_1.id), "model_version_id": str(self.model_version_1.id)},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"model_version_id": ["This ModelVersion is part of an archived model."]})

    def test_create_invalid_process_id(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": uuid.uuid4()}),
                data={"worker_version_id": str(self.version_1.id), "parents": []},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(response.json(), {"detail": "Not found."})

    def test_create_invalid_process_mode(self):
        self.client.force_login(self.user)

        for mode in set(ProcessMode) - {ProcessMode.Workers, ProcessMode.Dataset, ProcessMode.Local}:
            with self.subTest(mode=mode):
                self.process_2.mode = mode
                self.process_2.save()

                with self.assertNumQueries(6):
                    response = self.client.post(
                        reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                        {"worker_version_id": str(self.version_1.id)},
                        format="json",
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "process_id": ["WorkerRuns can only be created or updated on Workers or Dataset processes."],
                })

    def test_create_non_corpus_process_mode(self):
        process = self.user.processes.get(mode=ProcessMode.Local)
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(process.id)}),
                {"worker_version_id": str(self.version_1.id), "parents": []},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_create_process_already_started(self):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            farm=self.farm,
        )
        process.run()

        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(process.id)}),
                data={"worker_version_id": str(self.version_1.id), "parents": []},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "process_id": ["WorkerRuns cannot be added or updated on processes that have already started."],
        })

    def test_create(self):
        self.client.force_login(self.user)

        for mode in (ProcessMode.Workers, ProcessMode.Dataset):
            self.process_2.worker_runs.filter(version=self.version_1).delete()

            with self.subTest(mode=mode):
                self.process_2.mode = mode
                self.process_2.save()

                with self.assertNumQueries(8):
                    response = self.client.post(
                        reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                        {"worker_version_id": str(self.version_1.id), "parents": []},
                        format="json",
                    )
                    self.assertEqual(response.status_code, status.HTTP_201_CREATED)

                data = response.json()
                pk = data.pop("id")
                self.assertNotEqual(pk, self.run_1.id)
                self.assertDictEqual(data, {
                    "worker_version": {
                        "id": str(self.version_1.id),
                        "configuration": {"test": 42},
                        "docker_image": None,
                        "docker_image_iid": self.version_1.docker_image_iid,
                        "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                        "gpu_usage": "disabled",
                        "model_usage": FeatureUsage.Disabled.value,
                        "revision": {
                            "id": str(self.version_1.revision.id),
                            "author": "Test user",
                            "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                            "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                            "hash": "1337",
                            "message": "My w0rk3r",
                            "refs": []
                        },
                        "version": None,
                        "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                        "state": "available",
                        "worker": {
                            "id": str(self.worker_1.id),
                            "name": "Recognizer",
                            "slug": "reco",
                            "type": "recognizer"
                        }
                    },
                    "parents": [],
                    "model_version": None,
                    "configuration": None,
                    "process": {
                        "id": str(self.process_2.id),
                        "activity_state": "disabled",
                        "corpus": str(self.corpus.id),
                        "chunks": 1,
                        "mode": mode.value,
                        "name": None,
                        "state": "unscheduled",
                        "use_cache": False,
                    },
                    "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]}",
                })
                run = WorkerRun.objects.get(pk=pk)
                # Check generated summary, this supposes that self.version_1.revision does not have any git refs
                self.assertEqual(run.summary, f"Worker {self.worker_1.name} @ {str(self.version_1.id)[0:6]}")

    def test_create_via_repository_right(self):
        self.worker_1.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.worker_1.repository.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(self.version_1.id), "parents": []},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_empty(self):
        """
        The worker_version_id is required to create a worker run
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "worker_version_id": ["This field is required."],
        })

    def test_create_configuration(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={
                    "worker_version_id": str(self.version_1.id),
                    "parents": [],
                    "configuration_id": str(self.configuration_1.id)
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        pk = data.pop("id")
        self.assertNotEqual(pk, self.run_1.id)
        self.assertDictEqual(data, {
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "parents": [],
            "model_version": None,
            "configuration": {
                "id": str(self.configuration_1.id),
                "archived": False,
                "configuration": {"key": "value"},
                "name": "My config"
            },
            "process": {
                "id": str(self.process_2.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]} using configuration 'My config'",
        })
        run = WorkerRun.objects.get(pk=pk)
        # Check generated summary, this supposes that self.version_1.revision does not have any git refs
        self.assertEqual(run.summary, f"Worker {self.worker_1.name} @ {str(self.version_1.id)[0:6]} using configuration '{self.configuration_1.name}'")

    def test_create_invalid_configuration(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(self.version_1.id), "parents": [], "configuration_id": str(self.configuration_2.id)},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"configuration_id": ["The configuration must be part of the same worker."]})

    def test_list_summary_with_git_refs(self):
        """
        Create a worker run with a worker version related to a revision tagged with a GitRef
        """
        self.client.force_login(self.user)
        # Version with no git refs
        revision = Revision.objects.create(
            repo_id=self.repo.id,
            hash=uuid.uuid4().hex,
            message="Fake revision",
            author="Teklia Bot"
        )
        git_ref_names = ["main", "develop", "trunk", "master", "patate", "pouet"]
        expected_refs = []
        for git_ref_name in git_ref_names:
            new_ref = revision.refs.create(
                name=git_ref_name,
                type=GitRefType.Branch,
                repository=revision.repo
            )
            expected_refs.append({"id": str(new_ref.id), "name": git_ref_name, "type": "branch"})

        version = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=revision,
            configuration={},
            state=WorkerVersionState.Available,
            docker_image_iid=self.version_1.docker_image_iid,
        )

        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(version.id), "parents": []},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        pk = data.pop("id")
        self.assertNotEqual(pk, self.run_1.id)
        self.assertDictEqual(data, {
            "worker_version": {
                "id": str(version.id),
                "configuration": {},
                "docker_image": None,
                "docker_image_iid": version.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{version.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(version.revision.id),
                    "author": "Teklia Bot",
                    "commit_url": f"http://my_repo.fake/workers/worker/commit/{version.revision.hash}",
                    "created": version.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": version.revision.hash,
                    "message": "Fake revision",
                    "refs": expected_refs,
                },
                "version": None,
                "created": version.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "parents": [],
            "model_version": None,
            "configuration": None,
            "process": {
                "id": str(self.process_2.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
            },
            "summary": "Worker Recognizer @ main, develop, trunk, master, patate, pouet",
        })
        run = WorkerRun.objects.get(pk=pk)
        # Check generated summary
        self.assertEqual(run.summary, f"Worker {self.worker_1.name} @ main, develop, trunk, master, patate, pouet")

    def test_retrieve_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve_no_worker_execution_right(self):
        """
        A user can retrieve any worker run if they have an admin access to the process project
        """
        self.worker_1.memberships.update(level=Role.Guest.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": None,
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]}",
        })

    def test_retrieve_invalid_id(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:worker-run-details", kwargs={"pk": "12341234-1234-1234-1234-123412341234"})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": None,
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]}",
        })

    def test_retrieve_custom(self):
        self.client.force_login(self.user)
        # Update process attributes to ensure they do not cause extra requests
        page = self.corpus.elements.get(name="Volume 1, page 1r")
        self.local_process.element = page
        self.local_process.element_type = page.type
        self.local_process.folder_type = page.type
        self.local_process.prefix = "a"
        self.local_process.save()
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:worker-run-details", kwargs={"pk": str(self.run_custom.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.run_custom.id),
            "configuration": None,
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.local_process.id),
                "activity_state": "disabled",
                "corpus": None,
                "chunks": 1,
                "mode": "local",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": {
                    "id": str(page.id),
                    "type": "page",
                    "corpus": {
                        "id": str(self.corpus.id),
                        "name": "Unit Tests",
                        "public": True
                    },
                    "name": "Volume 1, page 1r",
                    "rotation_angle": 0,
                    "mirrored": False,
                    "thumbnail_url": None,
                    "zone": {
                        "id": str(page.id),
                        "image": {
                            "id": str(page.image.id),
                            "height": 1000,
                            "path": "img1",
                            "s3_url": None,
                            "server": {
                                "display_name": "Test Server",
                                "max_height": None,
                                "max_width": None,
                                "url": "http://server"
                            },
                            "status": "unchecked",
                            "url": "http://server/img1",
                            "width": 1000
                        },
                        "polygon": [[int(x), int(y)] for x, y in page.polygon],
                        "url": "http://server/img1/0,0,1000,1000/full/0/default.jpg",
                    },
                },
                "element_type": "page",
                "folder_type": "page",
                "prefix": "a",
            },
            "worker_version": {
                "id": str(self.version_custom.id),
                "configuration": {"custom": "value"},
                "docker_image": None,
                "docker_image_iid": None,
                "docker_image_name": None,
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": None,
                "version": 1,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "created",
                "worker": {
                    "id": str(self.worker_custom.id),
                    "name": "Custom worker",
                    "slug": "custom",
                    "type": "custom"
                }
            },
            "summary": "Worker Custom worker @ version 1",
        })

    @override_settings(PUBLIC_HOSTNAME="https://arkindex.localhost")
    def test_gitrefs_are_retrieved(self):
        """
        Check the GitRefs are retrieved with the revision
        """
        # Create gitrefs and check they were created
        commit_refs = [
            {"name": "refs/tags/v0.1.0", "type": "tag"},
            {"name": "refs/heads/branch1", "type": "branch"},
            {"name": "refs/heads/branch2", "type": "branch"},
        ]
        revision = Revision.objects.create(
            repo=self.repo,
            hash="1",
            message="commit message",
            author="bob",
        )

        for ref in commit_refs:
            GitRef.objects.create(repository=self.repo, revision=revision, name=ref["name"], type=ref["type"])

        refs = [
            {"name": ref.name, "type": ref.type, "repo": ref.repository}
            for ref in revision.refs.all()
        ]
        self.assertListEqual(refs, [
            {"name": "refs/tags/v0.1.0", "type": GitRefType.Tag, "repo": self.repo},
            {"name": "refs/heads/branch1", "type": GitRefType.Branch, "repo": self.repo},
            {"name": "refs/heads/branch2", "type": GitRefType.Branch, "repo": self.repo},
        ])
        self.assertEqual(self.repo.refs.count(), 3)

        # Assign the revision with gitrefs to the worker version
        self.version_1.revision = revision
        self.version_1.save()
        self.assertTrue(revision.refs.exists())

        # Check that the gitrefs are retrieved with RetrieveWorkerRun
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        version_revision = response.json()["worker_version"]["revision"]
        refs = version_revision.pop("refs")
        # Only asserting on the name and type of the refs since they were created in bulk using
        # GitLabProvider.update_or_create_ref, ignoring the IDs.
        refs_no_id = [
            {
                "name": item["name"],
                "type": item["type"]
            }
            for item in refs
        ]
        self.assertDictEqual(version_revision, {
            "author": "bob",
            "commit_url": "http://my_repo.fake/workers/worker/commit/1",
            "created": revision.created.isoformat().replace("+00:00", "Z"),
            "hash": "1",
            "id": str(revision.id),
            "message": "commit message",
        })
        self.assertCountEqual(refs_no_id, [
            {
                "name": "refs/tags/v0.1.0",
                "type": "tag"
            },
            {
                "name": "refs/heads/branch1",
                "type": "branch"
            },
            {
                "name": "refs/heads/branch2",
                "type": "branch"
            }
        ])

    def test_retrieve_local(self):
        """
        A user can retrieve a run on their own local process
        """
        run = self.local_process.worker_runs.create(version=self.version_1, parents=[])
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.get(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(run.id),
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "parents": [],
            "model_version": None,
            "configuration": None,
            "process": {
                "id": str(self.local_process.id),
                "activity_state": "disabled",
                "corpus": None,
                "chunks": 1,
                "mode": "local",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]}",
        })

    def test_retrieve_local_only_current_user(self):
        """
        A user cannot retrieve a run on another user's local process
        """
        run = WorkerRun.objects.filter(process__creator=self.superuser, process__mode=ProcessMode.Local).first()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve_agent(self):
        """
        A Ponos agent can retrieve a WorkerRun on a process where it has some assigned tasks
        """
        self.process_1.tasks.create(run=0, depth=0, slug="something", agent=self.agent)

        # Agent auth is not implemented in CE
        self.client.force_authenticate(user=self.agent)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.run_1.id),
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "parents": [],
            "model_version": None,
            "configuration": None,
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "mode": "workers",
                "chunks": 1,
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]}",
        })

    def test_retrieve_agent_unassigned(self):
        """
        A Ponos agent cannot retrieve a WorkerRun on a process where it does not have any assigned tasks
        """
        self.process_1.tasks.create(run=0, depth=0, slug="something", agent=None)

        # Agent auth is not implemented in CE
        self.client.force_authenticate(user=self.agent)
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_requires_nothing(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_update_requires_login(self):
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="commit message",
            author="bob",
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )

        with self.assertNumQueries(0):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": [str(run_2.id)],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.get_max_level", return_value=Role.Contributor.value)
    def test_update_no_project_admin_right(self, get_max_level_mock):
        """
        A user cannot update a worker run if they have no admin access on its process project
        """
        self.corpus.memberships.filter(user=self.user).update(level=Role.Contributor.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": []
                }
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have an admin access to this process."})

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    def test_update_invalid_process_mode(self):
        """
        A user cannot update a worker run on a local process
        """
        run = self.local_process.worker_runs.create(version=self.version_1, parents=[])
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                data={
                    "parents": []
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "process_id": ["WorkerRuns can only be created or updated on Workers or Dataset processes."],
        })

    def test_update_invalid_id(self):
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="commit message",
            author="bob",
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": "12341234-1234-1234-1234-123412341234"}),
                data={
                    "parents": [str(run_2.id)],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_nonexistent_parent(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": ["12341234-1234-1234-1234-123412341234"],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), [
            f"Can't add or update WorkerRun {self.run_1.id} because parents field isn't properly defined. It can be either because"
            " one or several UUIDs don't refer to existing WorkerRuns or either because listed WorkerRuns doesn't belong to the"
            " same Process than this WorkerRun."
        ])

    def test_update_duplicate_parents(self):
        self.client.force_login(self.user)
        run_2 = self.process_1.worker_runs.create(version=self.version_2)

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": [
                        str(run_2.id),
                        str(run_2.id),
                    ],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "parents": ["The parents of a WorkerRun must be unique."],
        })

    def test_duplicate_parents_signal(self):
        """
        Duplicates in WorkerRun.parents are also detected outside of the WorkerRun APIs
        """
        run_2 = self.process_1.worker_runs.create(
            version=self.version_2,
            parents=[],
        )

        run_2.parents = [self.run_1.id, self.run_1.id]
        with self.assertRaisesRegex(ValidationError, f"Can't add or update WorkerRun {run_2.id} because it has duplicate parents."):
            run_2.parents = [self.run_1.id, self.run_1.id]
            run_2.save()

    def test_update_process_id(self):
        """
        Process field cannot be updated
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "process_id": str(self.process_2.id),
                    "parents": [],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": None,
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]}",
        })
        self.run_1.refresh_from_db()
        self.assertEqual(self.run_1.process.id, self.process_1.id)

    def test_update_worker_version_id(self):
        """
        Version field cannot be updated
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "worker_version_id": str(self.version_2.id),
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": None,
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                },
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]}",
        })
        self.run_1.refresh_from_db()
        self.assertNotEqual(self.run_1.version_id, self.version_2.id)

    def test_update_configuration(self):
        self.client.force_login(self.user)
        self.assertEqual(self.run_1.configuration, None)
        # Check generated summary, before updating, it should not be that verbose
        self.assertEqual(self.run_1.summary, f"Worker Recognizer @ {str(self.run_1.version.id)[:6]}")

        with self.assertNumQueries(9):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": self.run_1.id}),
                data={
                    "parents": [],
                    "configuration_id": str(self.configuration_1.id)
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.run_1.refresh_from_db()

        self.assertEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": {
                "id": str(self.configuration_1.id),
                "archived": False,
                "configuration": {"key": "value"},
                "name": "My config"
            },
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]} using configuration 'My config'",
        })
        self.assertEqual(self.run_1.configuration.id, self.configuration_1.id)
        # Check generated summary, after the update, the configuration should be displayed as well
        self.assertEqual(self.run_1.summary, f"Worker {self.run_1.version.worker.name} @ {str(self.run_1.version.id)[:6]} using configuration '{self.configuration_1.name}'")

    def test_update_invalid_configuration(self):
        self.client.force_login(self.user)
        self.assertEqual(self.run_1.configuration, None)

        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": self.run_1.id}),
                data={"parents": [], "configuration_id": str(self.configuration_2.id)},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"configuration_id": ["The configuration must be part of the same worker."]})

    def test_update_process_already_started(self):
        """
        Update dependencies of a worker run is not possible once the process is started
        """
        self.process_1.run()
        self.assertTrue(self.process_1.tasks.exists())
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": [],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "process_id": ["WorkerRuns cannot be added or updated on processes that have already started."],
        })

    def test_update_model_version_not_allowed(self):
        """
        The model_version UUID is not allowed when the related version doesn't allow model_usage
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Disabled
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )

        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                data={
                    "model_version_id": str(self.model_version_1.id),
                    "parents": []
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": ["This worker version does not support models."]
        })

    def test_update_unknown_model_version(self):
        """
        Cannot use a model version id that doesn't exist
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )
        random_model_version_uuid = str(uuid.uuid4())

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                data={
                    "model_version_id": random_model_version_uuid,
                    "parents": []
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": [f'Invalid pk "{random_model_version_uuid}" - object does not exist.']
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=ModelVersion.objects.none())
    def test_update_model_version_no_access(self, filter_rights_mock):
        """
        Cannot update a worker run with a model_version UUID, when you don't have access to the model version
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )

        # Create a model version, the user has no access to
        model_no_access = Model.objects.create(name="Secret model")
        model_version_no_access = ModelVersion.objects.create(
            model=model_no_access,
            state=ModelVersionState.Available,
            size=8,
            hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            archive_hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
        )

        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                data={
                    "model_version_id": str(model_version_no_access.id),
                    "parents": []
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": [f'Invalid pk "{model_version_no_access.id}" - object does not exist.'],
        })

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Model, Role.Guest.value),
            call(self.user, Model, Role.Contributor.value),
        ])

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_update_model_version_guest(self, filter_rights_mock):
        """
        Cannot update a worker run with a model_version when you only have guest access to the model,
        and the model version has no tag or is not available
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )

        def filter_rights(user, model, level):
            """
            The filter_rights mock needs to return nothing when called for contributor access,
            and the models we will test on when called for guest access
            """
            if level == Role.Guest.value:
                return Model.objects.filter(id__in=(self.model_2.id, self.model_3.id))
            return Model.objects.none()

        filter_rights_mock.side_effect = filter_rights

        cases = [
            # On a model with a membership giving guest access
            (self.model_version_2, None, ModelVersionState.Created),
            (self.model_version_2, None, ModelVersionState.Error),
            (self.model_version_2, None, ModelVersionState.Available),
            (self.model_version_2, "blah", ModelVersionState.Created),
            (self.model_version_2, "blah", ModelVersionState.Error),
            # On a public model with no membership
            (self.model_version_3, None, ModelVersionState.Created),
            (self.model_version_3, None, ModelVersionState.Error),
            (self.model_version_3, None, ModelVersionState.Available),
            (self.model_version_3, "blah", ModelVersionState.Created),
            (self.model_version_3, "blah", ModelVersionState.Error),
        ]

        for model_version, tag, state in cases:
            filter_rights_mock.reset_mock()
            with self.subTest(model_version=model_version, tag=tag, state=state):
                model_version.tag = tag
                model_version.state = state
                model_version.save()

                with self.assertNumQueries(5):
                    response = self.client.put(
                        reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                        data={
                            "model_version_id": str(model_version.id),
                            "parents": [],
                        },
                        format="json",
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "model_version_id": [f'Invalid pk "{model_version.id}" - object does not exist.'],
                })

                self.assertListEqual(filter_rights_mock.call_args_list, [
                    call(self.user, Model, Role.Guest.value),
                    call(self.user, Model, Role.Contributor.value),
                ])

    def test_update_model_version_unavailable(self):
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run = self.process_1.worker_runs.create(
            version=version,
            parents=[],
        )
        self.model_version_1.state = ModelVersionState.Error
        self.model_version_1.save()

        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                data={
                    "model_version_id": str(self.model_version_1.id),
                    "parents": []
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": ["This ModelVersion is not in an Available state."]
        })

    def test_update_model_archived(self):
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run = self.process_1.worker_runs.create(version=version)
        self.model_1.archived = datetime.now(timezone.utc)
        self.model_1.save()

        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                data={
                    "model_version_id": str(self.model_version_1.id),
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": ["This ModelVersion is part of an archived model."],
        })

    def test_update_model_version_id(self):
        """
        Update the worker run by adding a model_version with a worker version that supports it
        """
        self.client.force_login(self.user)
        rev = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version_with_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Supported
        )
        run = self.process_1.worker_runs.create(
            version=version_with_model,
            parents=[],
        )
        self.assertEqual(run.model_version, None)
        # Check generated summary, before updating, there should be only information about the worker version
        self.assertEqual(run.summary, f"Worker {self.worker_1.name} @ {str(version_with_model.id)[:6]}")

        model_versions = [
            # Version on a model with contributor access
            self.model_version_1,
            # Available version with tag on a model with guest access
            self.model_version_2,
            # Available version with tag on a public model
            self.model_version_3,
        ]
        for model_version in model_versions:
            with self.subTest(model_version=model_version):
                with self.assertNumQueries(9):
                    response = self.client.put(
                        reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                        data={
                            "model_version_id": str(model_version.id),
                            "parents": [],
                        },
                        format="json",
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                run.refresh_from_db()
                self.assertEqual(response.json(), {
                    "id": str(run.id),
                    "configuration": None,
                    "model_version": {
                        "id": str(model_version.id),
                        "configuration": {},
                        "model": {
                            "id": str(model_version.model.id),
                            "name": model_version.model.name
                        },
                        "size": 8,
                        "state": "available",
                        "tag": model_version.tag,
                    },
                    "parents": [],
                    "process": {
                        "id": str(self.process_1.id),
                        "activity_state": "disabled",
                        "corpus": str(self.corpus.id),
                        "chunks": 1,
                        "mode": "workers",
                        "name": None,
                        "state": "unscheduled",
                        "use_cache": False,
                        "element": None,
                        "element_type": None,
                        "folder_type": None,
                        "prefix": None,
                    },
                    "worker_version": {
                        "id": str(version_with_model.id),
                        "configuration": {"test": "test2"},
                        "docker_image": None,
                        "docker_image_iid": None,
                        "docker_image_name": f"my_repo.fake/workers/worker/reco:{version_with_model.id}",
                        "gpu_usage": "disabled",
                        "model_usage": FeatureUsage.Supported.value,
                        "revision": {
                            "id": str(rev.id),
                            "author": "bob",
                            "commit_url": "http://my_repo.fake/workers/worker/commit/2",
                            "created": rev.created.isoformat().replace("+00:00", "Z"),
                            "hash": "2",
                            "message": "beep boop",
                            "refs": []
                        },
                        "version": None,
                        "created": version_with_model.created.isoformat().replace("+00:00", "Z"),
                        "state": "created",
                        "worker": {
                            "id": str(self.worker_1.id),
                            "name": "Recognizer",
                            "slug": "reco",
                            "type": "recognizer"
                        }
                    },
                    "summary": f"Worker Recognizer @ {str(version_with_model.id)[:6]} with model {model_version.model.name} @ {str(model_version.id)[:6]}",
                })
                self.assertEqual(run.model_version_id, model_version.id)
                self.assertEqual(run.summary, f"Worker {version_with_model.worker.name} @ {str(version_with_model.id)[:6]} with model {model_version.model.name} @ {str(model_version.id)[:6]}")

    def test_update_configuration_and_model_version(self):
        """
        Update the worker run by adding both a model_version and a worker configuration
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version_with_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run = self.process_1.worker_runs.create(
            version=version_with_model,
            parents=[],
        )
        self.assertIsNone(run.model_version)
        self.assertIsNone(run.configuration)
        # Check generated summary, before updating, there should be only information about the worker version
        self.assertEqual(run.summary, f"Worker {self.worker_1.name} @ {str(version_with_model.id)[:6]}")

        with self.assertNumQueries(10):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                data={
                    "model_version_id": str(self.model_version_1.id),
                    "configuration_id": str(self.configuration_1.id),
                    "parents": []
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        run.refresh_from_db()
        self.assertEqual(response.json(), {
            "id": str(run.id),
            "configuration": {
                "id": str(self.configuration_1.id),
                "archived": False,
                "configuration": {"key": "value"},
                "name": "My config"
            },
            "model_version": {
                "id": str(self.model_version_1.id),
                "configuration": {},
                "model": {
                    "id": str(self.model_1.id),
                    "name": "My model"
                },
                "size": 8,
                "state": "available",
                "tag": None
            },
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(version_with_model.id),
                "configuration": {"test": "test2"},
                "docker_image": None,
                "docker_image_iid": None,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{version_with_model.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Required.value,
                "revision": {
                    "id": str(rev_2.id),
                    "author": "bob",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/2",
                    "created": rev_2.created.isoformat().replace("+00:00", "Z"),
                    "hash": "2",
                    "message": "beep boop",
                    "refs": []
                },
                "version": None,
                "created": version_with_model.created.isoformat().replace("+00:00", "Z"),
                "state": "created",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "summary": f"Worker Recognizer @ {str(version_with_model.id)[:6]} with model My model @ {str(self.model_version_1.id)[:6]} using configuration 'My config'",
        })
        self.assertEqual(run.model_version_id, self.model_version_1.id)
        # Check generated summary, after updating, there should be information about the model loaded
        self.assertEqual(run.summary, f"Worker {self.worker_1.name} @ {str(version_with_model.id)[:6]} with model {self.model_version_1.model.name} @ {str(self.model_version_1.id)[:6]} using configuration '{self.configuration_1.name}'")

    def test_update(self):
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="commit message",
            author="bob",
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(10):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": [str(run_2.id)],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.run_1.refresh_from_db()
        self.assertDictEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": None,
            "model_version": None,
            "parents": [str(run_2.id)],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]}",
        })

    def test_update_unique(self):
        self.client.force_login(self.user)
        self.version_1.model_usage = FeatureUsage.Required
        self.version_1.save()
        cases = [
            (None, None),
            (None, self.configuration_1),
            (self.model_version_1, None),
            (self.model_version_1, self.configuration_1),
        ]
        for model_version, configuration in cases:
            with self.subTest(model_version=model_version, configuration=configuration):
                # Erase any previous failures
                self.process_1.worker_runs.exclude(id=self.run_1.id).delete()

                self.run_1.model_version = model_version
                self.run_1.configuration = configuration
                self.run_1.save()

                # Ensure the other run has different values before updating to avoid conflicts
                run_2 = self.process_1.worker_runs.create(
                    version=self.version_1,
                    model_version=None if model_version else self.model_version_1,
                    configuration=None if configuration else self.configuration_1,
                )

                # Having a model version or a configuration adds one query for each
                query_count = 5 + bool(model_version) + bool(configuration)

                with self.assertNumQueries(query_count):
                    response = self.client.put(
                        reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                        data={
                            # Update the second worker run to the first worker run's values to cause a conflict
                            "model_version_id": str(model_version.id) if model_version else None,
                            "configuration_id": str(configuration.id) if configuration else None,
                        },
                        format="json",
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "__all__": ["A WorkerRun already exists on this process with the selected worker version, model version and configuration."],
                })

    def test_update_agent(self):
        """
        Ponos agents cannot update WorkerRuns, even when they can access them
        """
        self.process_1.tasks.create(run=0, depth=0, slug="something", agent=self.agent)

        # Agent auth is not implemented in CE
        self.client.force_authenticate(user=self.agent)
        with self.assertNumQueries(2):
            response = self.client.put(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {
            "detail": "You do not have an admin access to this process.",
        })

    def test_partial_update_requires_login(self):
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="commit message",
            author="bob",
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )

        with self.assertNumQueries(0):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": [str(run_2.id)],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.get_max_level", return_value=Role.Contributor.value)
    def test_partial_update_no_project_admin_right(self, get_max_level_mock):
        """
        A user cannot update a worker run if they have no admin access on its process project
        """
        self.corpus.memberships.filter(user=self.user).update(level=Role.Contributor.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have an admin access to this process."})

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    def test_partial_update_invalid_id(self):
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="commit message",
            author="bob",
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )

        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": "12341234-1234-1234-1234-123412341234"}),
                data={
                    "parents": [str(run_2.id)],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_partial_update_local(self):
        """
        A user cannot update a worker run on a local process
        """
        run = self.local_process.worker_runs.create(version=self.version_1, parents=[])
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                data={
                    "parents": []
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "process_id": ["WorkerRuns can only be created or updated on Workers or Dataset processes."],
        })

    def test_partial_update_nonexistent_parent(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": ["12341234-1234-1234-1234-123412341234"],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), [
            f"Can't add or update WorkerRun {self.run_1.id} because parents field isn't properly defined. It can be either because"
            " one or several UUIDs don't refer to existing WorkerRuns or either because listed WorkerRuns doesn't belong to the"
            " same Process than this WorkerRun."
        ])

    def test_partial_update_process_id(self):
        """
        Process field cannot be updated
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "process_id": str(self.process_2.id),
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": None,
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]}",
        })
        self.run_1.refresh_from_db()
        self.assertEqual(self.run_1.process.id, self.process_1.id)

    def test_partial_update_worker_version_id(self):
        """
        Version field cannot be updated
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "worker_version_id": str(self.version_2.id),
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": None,
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]}",
        })
        self.run_1.refresh_from_db()
        self.assertNotEqual(self.run_1.version_id, self.version_2.id)

    def test_partial_update_configuration(self):
        self.client.force_login(self.user)
        self.assertEqual(self.run_1.configuration, None)
        self.assertEqual(self.run_1.summary, f"Worker {self.run_1.version.worker.name} @ {str(self.run_1.version.id)[:6]}")

        with self.assertNumQueries(9):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": self.run_1.id}),
                data={
                    "configuration_id": str(self.configuration_1.id)
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.run_1.refresh_from_db()
        self.assertEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": {
                "id": str(self.configuration_1.id),
                "archived": False,
                "configuration": {"key": "value"},
                "name": "My config"
            },
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]} using configuration 'My config'",
        })
        self.assertEqual(self.run_1.configuration.id, self.configuration_1.id)
        self.assertEqual(self.run_1.summary, f"Worker {self.run_1.version.worker.name} @ {str(self.run_1.version.id)[:6]} using configuration '{self.configuration_1.name}'")

    def test_partial_update_invalid_configuration(self):
        self.client.force_login(self.user)
        self.assertEqual(self.run_1.configuration, None)

        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": self.run_1.id}),
                data={"configuration_id": str(self.configuration_2.id)},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"configuration_id": ["The configuration must be part of the same worker."]})

    def test_partial_update_process_already_started(self):
        """
        Update dependencies of a worker run is not possible once the process is started
        """
        self.process_1.run()
        self.assertTrue(self.process_1.tasks.exists())
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": [],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "process_id": ["WorkerRuns cannot be added or updated on processes that have already started."],
        })

    def test_partial_update_model_version_not_allowed(self):
        """
        The model_version UUID is not allowed when the related version doesn't allow model_usage
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Disabled
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )

        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                data={
                    "model_version_id": str(self.model_version_1.id),
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": ["This worker version does not support models."]
        })

    def test_partial_update_unknown_model_version(self):
        """
        Cannot use a model version id that doesn't exist
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )
        random_model_version_uuid = str(uuid.uuid4())

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                data={
                    "model_version_id": random_model_version_uuid,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": [f'Invalid pk "{random_model_version_uuid}" - object does not exist.']
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=ModelVersion.objects.none())
    def test_partial_update_model_version_no_access(self, filter_rights_mock):
        """
        Cannot update a worker run with a model_version UUID, when you don't have access to the model version
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )

        # Create a model version, the user has no access to
        model_no_access = Model.objects.create(name="Secret model")
        model_version_no_access = ModelVersion.objects.create(model=model_no_access, state=ModelVersionState.Available, size=8, hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", archive_hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb")

        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                data={
                    "model_version_id": str(model_version_no_access.id),
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": [f'Invalid pk "{model_version_no_access.id}" - object does not exist.'],
        })

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Model, Role.Guest.value),
            call(self.user, Model, Role.Contributor.value),
        ])

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_partial_update_model_version_guest(self, filter_rights_mock):
        """
        Cannot update a worker run with a model_version when you only have guest access to the model,
        and the model version has no tag or is not available
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )

        def filter_rights(user, model, level):
            """
            The filter_rights mock needs to return nothing when called for contributor access,
            and the models we will test on when called for guest access
            """
            if level == Role.Guest.value:
                return Model.objects.filter(id__in=(self.model_2.id, self.model_3.id))
            return Model.objects.none()

        filter_rights_mock.side_effect = filter_rights

        cases = [
            # On a model with a membership giving guest access
            (self.model_version_2, None, ModelVersionState.Created),
            (self.model_version_2, None, ModelVersionState.Error),
            (self.model_version_2, None, ModelVersionState.Available),
            (self.model_version_2, "blah", ModelVersionState.Created),
            (self.model_version_2, "blah", ModelVersionState.Error),
            # On a public model with no membership
            (self.model_version_3, None, ModelVersionState.Created),
            (self.model_version_3, None, ModelVersionState.Error),
            (self.model_version_3, None, ModelVersionState.Available),
            (self.model_version_3, "blah", ModelVersionState.Created),
            (self.model_version_3, "blah", ModelVersionState.Error),
        ]

        for model_version, tag, state in cases:
            filter_rights_mock.reset_mock()
            with self.subTest(model_version=model_version, tag=tag, state=state):
                model_version.tag = tag
                model_version.state = state
                model_version.save()

                with self.assertNumQueries(5):
                    response = self.client.patch(
                        reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                        data={
                            "model_version_id": str(model_version.id),
                        },
                        format="json",
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "model_version_id": [f'Invalid pk "{model_version.id}" - object does not exist.'],
                })

                self.assertListEqual(filter_rights_mock.call_args_list, [
                    call(self.user, Model, Role.Guest.value),
                    call(self.user, Model, Role.Contributor.value),
                ])

    def test_partial_update_model_version_unavailable(self):
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run = self.process_1.worker_runs.create(
            version=version,
            parents=[],
        )
        self.model_version_1.state = ModelVersionState.Error
        self.model_version_1.save()

        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                data={
                    "model_version_id": str(self.model_version_1.id),
                    "parents": []
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": ["This ModelVersion is not in an Available state."],
        })

    def test_partial_update_model_archived(self):
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run = self.process_1.worker_runs.create(version=version)
        self.model_1.archived = datetime.now(timezone.utc)
        self.model_1.save()

        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                data={
                    "model_version_id": str(self.model_version_1.id),
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": ["This ModelVersion is part of an archived model."],
        })

    def test_partial_update_model_version(self):
        """
        Update the worker run by adding a model_version with a worker version that supports it
        """
        self.client.force_login(self.user)
        rev = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version_with_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run = self.process_1.worker_runs.create(
            version=version_with_model,
            parents=[],
        )
        self.assertIsNone(run.model_version_id)
        self.assertEqual(run.summary, f"Worker Recognizer @ {str(version_with_model.id)[:6]}")

        model_versions = [
            # Version on a model with contributor access
            self.model_version_1,
            # Available version with tag on a model with guest access
            self.model_version_2,
            # Available version with tag on a public model
            self.model_version_3,
        ]
        for model_version in model_versions:
            with self.subTest(model_version=model_version):
                with self.assertNumQueries(9):
                    response = self.client.patch(
                        reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                        data={
                            "model_version_id": str(model_version.id),
                        },
                        format="json",
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                run.refresh_from_db()
                self.assertEqual(response.json(), {
                    "id": str(run.id),
                    "configuration": None,
                    "model_version": {
                        "id": str(model_version.id),
                        "configuration": {},
                        "model": {
                            "id": str(model_version.model.id),
                            "name": model_version.model.name
                        },
                        "size": 8,
                        "state": "available",
                        "tag": model_version.tag,
                    },
                    "parents": [],
                    "process": {
                        "id": str(self.process_1.id),
                        "activity_state": "disabled",
                        "corpus": str(self.corpus.id),
                        "chunks": 1,
                        "mode": "workers",
                        "name": None,
                        "state": "unscheduled",
                        "use_cache": False,
                        "element": None,
                        "element_type": None,
                        "folder_type": None,
                        "prefix": None,
                    },
                    "worker_version": {
                        "id": str(version_with_model.id),
                        "configuration": {"test": "test2"},
                        "docker_image": None,
                        "docker_image_iid": None,
                        "docker_image_name": f"my_repo.fake/workers/worker/reco:{version_with_model.id}",
                        "gpu_usage": "disabled",
                        "model_usage": FeatureUsage.Required.value,
                        "revision": {
                            "id": str(rev.id),
                            "author": "bob",
                            "commit_url": "http://my_repo.fake/workers/worker/commit/2",
                            "created": rev.created.isoformat().replace("+00:00", "Z"),
                            "hash": "2",
                            "message": "beep boop",
                            "refs": []
                        },
                        "version": None,
                        "created": version_with_model.created.isoformat().replace("+00:00", "Z"),
                        "state": "created",
                        "worker": {
                            "id": str(self.worker_1.id),
                            "name": "Recognizer",
                            "slug": "reco",
                            "type": "recognizer"
                        }
                    },
                    "summary": f"Worker Recognizer @ {str(version_with_model.id)[:6]} with model {model_version.model.name} @ {str(model_version.id)[:6]}",
                })
                self.assertEqual(run.model_version_id, model_version.id)
                self.assertEqual(run.summary, f"Worker {version_with_model.worker.name} @ {str(version_with_model.id)[:6]} with model {model_version.model.name} @ {str(model_version.id)[:6]}")

    def test_partial_update_model_version_with_configuration(self):
        """
        Updating the worker run by adding a model_version with a worker version
        doesn't erase previously loaded worker configuration
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="beep boop",
            author="bob",
        )
        version_with_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run = self.process_1.worker_runs.create(
            version=version_with_model,
            parents=[],
            configuration=self.configuration_1
        )
        self.assertEqual(run.model_version_id, None)

        with self.assertNumQueries(9):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                data={
                    "model_version_id": str(self.model_version_1.id),
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        run.refresh_from_db()
        self.assertEqual(response.json(), {
            "id": str(run.id),
            "configuration": {
                "archived": False,
                "configuration": {"key": "value"},
                "id": str(self.configuration_1.id),
                "name": "My config"
            },
            "model_version": {
                "id": str(self.model_version_1.id),
                "configuration": {},
                "model": {
                    "id": str(self.model_1.id),
                    "name": "My model"
                },
                "size": 8,
                "state": "available",
                "tag": None
            },
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(version_with_model.id),
                "configuration": {"test": "test2"},
                "docker_image": None,
                "docker_image_iid": None,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{version_with_model.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Required.value,
                "revision": {
                    "id": str(rev_2.id),
                    "author": "bob",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/2",
                    "created": rev_2.created.isoformat().replace("+00:00", "Z"),
                    "hash": "2",
                    "message": "beep boop",
                    "refs": []
                },
                "version": None,
                "created": version_with_model.created.isoformat().replace("+00:00", "Z"),
                "state": "created",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "summary": f"Worker Recognizer @ {str(version_with_model.id)[:6]} with model My model @ {str(self.model_version_1.id)[:6]} using configuration 'My config'",
        })
        self.assertEqual(run.model_version_id, self.model_version_1.id)
        self.assertEqual(run.configuration_id, self.configuration_1.id)

    def test_partial_update(self):
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="commit message",
            author="bob",
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(10):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": [str(run_2.id)],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.run_1.refresh_from_db()
        self.assertDictEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": None,
            "model_version": None,
            "parents": [str(run_2.id)],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image": None,
                "docker_image_iid": self.version_1.docker_image_iid,
                "docker_image_name": f"my_repo.fake/workers/worker/reco:{self.version_1.id}",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision": {
                    "id": str(self.version_1.revision.id),
                    "author": "Test user",
                    "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                    "created": self.version_1.revision.created.isoformat().replace("+00:00", "Z"),
                    "hash": "1337",
                    "message": "My w0rk3r",
                    "refs": []
                },
                "version": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "summary": f"Worker Recognizer @ {str(self.version_1.id)[:6]}",
        })

    def test_partial_update_unique(self):
        self.client.force_login(self.user)
        self.version_1.model_usage = FeatureUsage.Required
        self.version_1.save()
        cases = [
            (None, None),
            (None, self.configuration_1),
            (self.model_version_1, None),
            (self.model_version_1, self.configuration_1),
        ]
        for model_version, configuration in cases:
            with self.subTest(model_version=model_version, configuration=configuration):
                # Erase any previous failures
                self.process_1.worker_runs.exclude(id=self.run_1.id).delete()

                self.run_1.model_version = model_version
                self.run_1.configuration = configuration
                self.run_1.save()

                # Ensure the other run has different values before updating to avoid conflicts
                run_2 = self.process_1.worker_runs.create(
                    version=self.version_1,
                    model_version=None if model_version else self.model_version_1,
                    configuration=None if configuration else self.configuration_1,
                )

                # Having a model version or a configuration adds one query for each
                query_count = 5 + bool(model_version) + bool(configuration)

                with self.assertNumQueries(query_count):
                    response = self.client.patch(
                        reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                        data={
                            # Update the second worker run to the first worker run's values to cause a conflict
                            "model_version_id": str(model_version.id) if model_version else None,
                            "configuration_id": str(configuration.id) if configuration else None,
                        },
                        format="json",
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "__all__": ["A WorkerRun already exists on this process with the selected worker version, model version and configuration."],
                })

    def test_partial_update_agent(self):
        """
        Ponos agents cannot update WorkerRuns, even when they can access them
        """
        self.process_1.tasks.create(run=0, depth=0, slug="something", agent=self.agent)

        # Agent auth is not implemented in CE
        self.client.force_authenticate(user=self.agent)
        with self.assertNumQueries(2):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {
            "detail": "You do not have an admin access to this process.",
        })

    def test_delete_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_invalid_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": "12341234-1234-1234-1234-123412341234"})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_delete_no_worker_execution_right(self):
        """
        A user can delete a worker run with no right on its worker but an admin access to the process project
        """
        self.worker_1.memberships.update(level=Role.Guest.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        with self.assertRaises(WorkerRun.DoesNotExist):
            self.run_1.refresh_from_db()

    def test_delete_local(self):
        """
        A user cannot delete a worker run on a local process
        """
        run = self.local_process.worker_runs.create(version=self.version_1, parents=[])
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), ["WorkerRuns can only be deleted from Workers or Dataset processes."])

    def test_delete(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        with self.assertRaises(WorkerRun.DoesNotExist):
            self.run_1.refresh_from_db()

    def test_delete_with_children(self):
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="commit message",
            author="bob",
        )
        rev_3 = self.repo.revisions.create(
            hash="3",
            message="commit message",
            author="dylan",
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        version_3 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_3,
            configuration={"test": "test3"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[self.run_1.id],
        )
        run_3 = self.process_1.worker_runs.create(
            version=version_3,
            parents=[self.run_1.id, run_2.id],
        )

        self.assertTrue(self.run_1.id in run_2.parents)
        self.assertTrue(self.run_1.id in run_3.parents)
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        with self.assertRaises(WorkerRun.DoesNotExist):
            self.run_1.refresh_from_db()
        run_2.refresh_from_db()
        run_3.refresh_from_db()
        self.assertEqual(run_2.parents, [])
        self.assertEqual(run_3.parents, [run_2.id])

    def test_delete_started_process(self):
        """
        A user shouldn't be able to delete the worker run of a started process
        """
        self.client.force_login(self.user)
        self.process_1.run()
        self.process_1.tasks.update(state=State.Running)

        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), ["WorkerRuns cannot be deleted from a process that has already started."])

    def test_delete_agent(self):
        """
        Ponos agents cannot delete WorkerRuns, even when they can access them
        """
        self.process_1.tasks.create(run=0, depth=0, slug="something", agent=self.agent)

        # Agent auth is not implemented in CE
        self.client.force_authenticate(user=self.agent)
        with self.assertNumQueries(1):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {
            "detail": "You do not have an admin access to this process.",
        })

    def test_build_task_no_parent(self):
        task, parent_slugs = self.run_1.build_task(self.process_1, ENV.copy(), "import", "/data/import/elements.json")

        self.assertEqual(task.slug, f"reco_{str(self.run_1.id)[0:6]}")
        self.assertEqual(task.image, self.version_1.docker_image_iid)
        self.assertEqual(task.command, None)
        self.assertEqual(task.shm_size, None)
        self.assertEqual(parent_slugs, ["import"])
        self.assertEqual(task.env, {
            "ARKINDEX_PROCESS_ID": "12345",
            "ARKINDEX_TASK_TOKEN": str(task.token),
            "TASK_ELEMENTS": "/data/import/elements.json",
            "ARKINDEX_WORKER_RUN_ID": str(self.run_1.id),
        })

    def test_build_task_with_chunk(self):
        task, parent_slugs = self.run_1.build_task(self.process_1, ENV.copy(), "import", "/data/import/elements.json", chunk=4)

        self.assertEqual(task.slug, f"reco_{str(self.run_1.id)[0:6]}_4")
        self.assertEqual(task.image, self.version_1.docker_image_iid)
        self.assertEqual(task.command, None)
        self.assertEqual(task.shm_size, None)
        self.assertEqual(parent_slugs, ["import"])
        self.assertEqual(task.env, {
            "ARKINDEX_PROCESS_ID": "12345",
            "ARKINDEX_TASK_TOKEN": str(task.token),
            "ARKINDEX_TASK_CHUNK": "4",
            "TASK_ELEMENTS": "/data/import/elements.json",
            "ARKINDEX_WORKER_RUN_ID": str(self.run_1.id),
        })

    def test_build_task_with_parent(self):
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="commit message",
            author="bob",
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        version_2.docker_image_iid = "evaunit:latest"
        version_2.state = WorkerVersionState.Available
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[self.run_1.id],
        )

        task, parent_slugs = run_2.build_task(self.process_1, ENV.copy(), "import", "/data/import/elements.json")

        self.assertEqual(task.slug, f"reco_{str(run_2.id)[0:6]}")
        self.assertEqual(task.image, version_2.docker_image_iid)
        self.assertEqual(task.command, None)
        self.assertEqual(task.shm_size, None)
        self.assertEqual(parent_slugs, [f"reco_{str(self.run_1.id)[0:6]}"])
        self.assertEqual(task.env, {
            "ARKINDEX_PROCESS_ID": "12345",
            "ARKINDEX_TASK_TOKEN": str(task.token),
            "TASK_ELEMENTS": "/data/import/elements.json",
            "ARKINDEX_WORKER_RUN_ID": str(run_2.id),
        })

    def test_build_task_with_parent_and_chunk(self):
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="commit message",
            author="bob",
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        version_2.docker_image_iid = "evaunit:latest"
        version_2.state = WorkerVersionState.Available
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[self.run_1.id],
        )

        task, parent_slugs = run_2.build_task(self.process_1, ENV.copy(), "import", "/data/import/elements.json", chunk=4)

        self.assertEqual(task.slug, f"reco_{str(run_2.id)[0:6]}_4")
        self.assertEqual(task.image, version_2.docker_image_iid)
        self.assertEqual(task.command, None)
        self.assertEqual(task.shm_size, None)
        self.assertEqual(parent_slugs, [f"reco_{str(self.run_1.id)[0:6]}_4"])
        self.assertEqual(task.env, {
            "ARKINDEX_PROCESS_ID": "12345",
            "ARKINDEX_TASK_TOKEN": str(task.token),
            "ARKINDEX_TASK_CHUNK": "4",
            "TASK_ELEMENTS": "/data/import/elements.json",
            "ARKINDEX_WORKER_RUN_ID": str(run_2.id),
        })

    def test_build_task_shm_size(self):
        self.version_1.configuration = {
            "docker": {
                "shm_size": 505,
            }
        }
        task, parent_slugs = self.run_1.build_task(self.process_1, ENV.copy(), "import", "/data/import/elements.json")

        self.assertEqual(task.slug, f"reco_{str(self.run_1.id)[0:6]}")
        self.assertEqual(task.image, self.version_1.docker_image_iid)
        self.assertEqual(task.command, None)
        self.assertEqual(task.shm_size, 505)
        self.assertEqual(parent_slugs, ["import"])
        self.assertEqual(task.env, {
            "ARKINDEX_PROCESS_ID": "12345",
            "ARKINDEX_TASK_TOKEN": str(task.token),
            "TASK_ELEMENTS": "/data/import/elements.json",
            "ARKINDEX_WORKER_RUN_ID": str(self.run_1.id),
        })

    def test_build_task_unavailable_version(self):
        rev_2 = self.repo.revisions.create(
            hash="2",
            message="commit message",
            author="bob",
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        version_2.docker_image_iid = "evaunit:latest"
        self.assertEqual(version_2.state, WorkerVersionState.Created)
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[self.run_1.id],
        )

        with self.assertRaisesRegex(
            AssertionError,
            f"Worker Version {version_2.id} is not available and cannot be used to build a task."
        ):
            run_2.build_task(self.process_1, ENV.copy(), "import", "/data/import/elements.json")

    def test_build_task_unavailable_model_version(self):
        self.model_version_1.state = ModelVersionState.Created
        self.model_version_1.save()
        self.run_1.model_version = self.model_version_1
        self.run_1.save()
        with self.assertRaisesRegex(
            AssertionError,
            f"ModelVersion {self.model_version_1.id} is not available and cannot be used to build a task."
        ):
            self.run_1.build_task(self.process_1, ENV.copy(), "import", "/data/import/elements.json")
